// eigen_solvers_wrappers.cc -*-c++-*-

#include<externals_config_0.h>

#define ARPACK_DNAUPD PR1_FORTRAN_NAME(dnaupd)
#define ARPACK_DNEUPD PR1_FORTRAN_NAME(dneupd)

extern "C" {
  void
  ARPACK_DNAUPD(int *const ido, char const *const bmat, int const *const n,
		char const *const which, int const *const nev,
		double const *const tol, double *const resid,
		int const *const ncv, double *const v,int const *const ldv,
		int const *const iparam, int *const ipntr,
		double *const workd, double *const workl, int *const lworkl,
		int *const info);
  //ncv >= 2*nev is reasonable ( ARPACK user guide, p.30 )

  void
  ARPACK_DNEUPD(int const *const rvec, char const *const howmny,
		int *const select, double *const dr, double *const di,
		double *const z,int const *const ldz,
		double const *const sigmar, double const *const sigmai,
		double *const workev,
		char const *const bmat, int const *const n,
		char const *const which, int const *const nev,
		double const *const tol, double *const resid,
		int const *const ncv, double *const v,int const *const ldv,
		int const *const iparam, int *const ipntr,
		double *const workd, double *const workl, int *const lworkl,
		int *const info);
}

#include<cassert>
#include<iostream>
#include<cstdlib>
#include<algorithm>

//#include <boost/numeric/bindings/umfpack/umfpack.hpp> //for operation codes
// workaround
#define UMFPACK_Pt_L    (3)     /* P'Lx=b  */
#define UMFPACK_U_Qt    (9)     /* UQ'x=b  */

#include <boost/numeric/bindings/traits/std_vector.hpp>
#include <boost/numeric/bindings/traits/ublas_vector.hpp>
#include <boost/numeric/bindings/traits/ublas_matrix.hpp>

#include<boost/numeric/ublas/vector_proxy.hpp>
#include<boost/numeric/ublas/operation.hpp>

#include<eigen_solvers_wrappers.h>

//#include<GlobalTimer.h> // debug\log\profile
//#include<base/logstream.h> // debug\log\profile

namespace pu{

  LAL_TypeEigDriverNonSym<double>::
  LAL_TypeEigDriverNonSym(boost::shared_ptr<matrix_t const> const &pMi,
			  boost::shared_ptr<matrix_t const> const &pKi)
    :factorized(*pMi),pK(pKi),
     bmat('I'),n(pMi->size1())
  {
    assert(pMi->size1() == pMi->size2());
    assert(pMi->size1() == pKi->size1());
    assert(pMi->size1() == pKi->size2());
  }

  void LAL_TypeEigDriverNonSym<double>::
  solve(double const tol_in,std::string const &which_in,
	unsigned const nev_in){
    ido = 0;
    which = which_in;
    assert(n>=0);
    // -2 , otherwise dnaupd issues error
    nev = std::min(nev_in,static_cast<unsigned>(n)-2);
    tol = tol_in;
    resid.resize(n,false);

    // 2 reasonabe ratio  - from user guide.
    ncv = std::min(n,std::max(nev+2,4*nev+nev/2)); //4+1/2 ratio
    v.resize(n,ncv,false);
    ldv = n; // v.size1()

    iparam[0] = 1; //exact shifts - p. 18 of ARPACK user guide
    const int maxiter(30);
    iparam[2] = maxiter; // maximum number of iterations
    iparam[3] = 1; // said to do so,see dnaupd.f
    iparam[6] = 1; // regular mode

    workd.resize(3*n,false);
    lworkl = 3*ncv*ncv + 6*ncv; // at least 
    workl.resize(lworkl);

    info = 0; // no input vector

    namespace traits = boost::numeric::bindings::traits;
    boost::numeric::ublas::vector<double>
      mult_work_space1(n),
      mult_work_space2(n);

    // debug/profile/log
    //deallog << GlobalTimer.elapsed()
    //    <<" sec: Start eigenvalues lookup process." << std::endl;

    // debug/profile/log
    unsigned mult_count(0);

    while(true){
      ARPACK_DNAUPD(&ido,&bmat,&n,which.c_str(),&nev,&tol,
		    traits::vector_storage(resid),&ncv,
		    traits::matrix_storage(v),&ldv,
		    traits::vector_storage(iparam),
		    traits::vector_storage(ipntr),
		    traits::vector_storage(workd),
		    traits::vector_storage(workl),&lworkl,&info);
      if(ido == 99){ // end condition
	break;	    
      }
      switch(ido){
      case 1 : case -1:

	// debug/profile/log
	++mult_count;
	//deallog << GlobalTimer.elapsed()
	//<<" sec: Start Y=Op*X. " << mult_count << std::endl;

	typedef boost::numeric::ublas::vector<double> vector_t;
	typedef boost::numeric::ublas::vector_range<vector_t> range_t;
	typedef boost::numeric::ublas::range range;
	factorized.mldivide(UMFPACK_U_Qt,
			    range_t(workd,range(ipntr[0],ipntr[0]+n)),
			    mult_work_space1);
	boost::numeric::ublas::axpy_prod(*pK,mult_work_space1,
					 mult_work_space2,true);
	factorized.scale(mult_work_space2,mult_work_space1);
	{
	  range_t mult_target(workd,range(ipntr[1],ipntr[1]+n));
	  factorized.mldivide(UMFPACK_Pt_L,mult_work_space1,mult_target);
	}
	// debug/profile/log
	//deallog << GlobalTimer.elapsed()
	//<<" sec: End Y=Op*X. " << mult_count << std::endl;
	break;
      default:
	std::cerr <<"Wrong arpack::dnaupd ido communicator = "<<ido<<std::endl;
	std::exit(-1);
      }
    }

    // debug/profile/log
    //deallog << GlobalTimer.elapsed()
    //    <<" sec: End eigenvalues lookup process, total "
    //    << mult_count << " Y=OP*X multiplications." << std::endl;
    if (info < 0) {
      std::cerr << "Error with arpack::dnaupd, info = " << info
		<< "\nCheck the documentation of this subroutine." <<std::endl;
      std::exit(-1);
    }
    if (info == 1){
      std::cerr << "Warning with arpack::dnaupd: max iterations reached ("
		<<maxiter<< "),\nonly " <<iparam[4]<< " of requested " << nev
		<< " eigenvalues converged." << std::endl;
    }
  }

  void LAL_TypeEigDriverNonSym<double>::
  get_all_converged_eigs(std::vector<std::complex<double> > &result){
    // assumption: iparam[4] (number of converged) <= nev
    result.reserve(iparam[4]);
    namespace traits = boost::numeric::bindings::traits;
    int const rvec(false);
    char const hownmy('A');
    std::vector<int> select(ncv,true);
    boost::numeric::ublas::vector<double> dr(nev+1),di(nev+1);
    boost::numeric::ublas::matrix<double> z;
    int const ldz(0);

    double const sigmar(0),sigmai(0);
    //should not be referenced by dneupd in mode 1

    boost::numeric::ublas::vector<double> workev(3*ncv);
    
    ARPACK_DNEUPD(&rvec,&hownmy,traits::vector_storage(select),
		  traits::vector_storage(dr),traits::vector_storage(di),
		  traits::matrix_storage(z),&ldz,&sigmar,&sigmai,
		  traits::vector_storage(workev),
		  &bmat,&n,which.c_str(),&nev,&tol,
		  traits::vector_storage(resid),&ncv,
		  traits::matrix_storage(v),&ldv,
		  traits::vector_storage(iparam),
		  traits::vector_storage(ipntr),traits::vector_storage(workd),
		  traits::vector_storage(workl),&lworkl,&info);
    if(info != 0){
      std::cerr << "Unnormal exit of arpack::dneupd, info = "
		<< info << std::endl;
      std::exit(-1);
    }
    for(/*unsigned*/int i(0);i<iparam[4];++i){
      result.push_back(std::complex<double>(dr[i],di[i]));
    }
  }
}//of pu
